"""
Author Aurore Guillevic
email  aurore.guillevic@inria.fr
date   March 1st, 2019

Example of Quadratic Sieve in Z[i] "à la" NFS for p=1109.
(Coppersmith Odlyzko Schroeppel)
"""
# init.
# p = 1 mod 4 so that f=x^2+1 has a root mod p, in other words, p=u^2+v^2.
#p = 1000
#while not ((p % 4) == 1 and p.is_prime() and ((p-1)//4).is_prime()):
#    p = next_prime(p)

from examples_pretty_print import pretty_print_product, pretty_print_tex_product

def is_smooth(n,B):
    """ returns True if n is B-smooth = product of primes up to B (included)"""
    fac_n = ZZ(n).factor()
    smooth_n = True
    for (l,_) in fac_n:
        if l > B:
            return False
    return True

p = 1109
r = (p-1)//4
Fr = GF(r) # r = (p-1)/4
print("p = {}".format(p))
print("(p-1)/4 = {}".format(r))

Fp = GF(p)
sqrt_1 = sqrt(Fp(-1))
M = Matrix(ZZ, 2,2, [p, 0, ZZ(sqrt_1), 1])
R = M.LLL()
u = R[0][0]
v = R[0][1]

print("{}^2 + {}^2 = {}, p = {}".format(u,v, u^2+v^2, p))

u=22
v=25
# u/v
# polynomials: f0 = x^2+1, f1 = v*x-u
ZZx.<x> = ZZ[]
f1 = x^2+1
f0 = v*x-u
B = 13 # smoothness bound
A = 21 # bound on max(|a|,|b|)
print("factor basis f=x^2+1")
for l in prime_range(2, B+1):
    Fl = GF(l)
    Flz.<z> = Fl[]
    fl = Flz(f1)
    if len(fl.roots())>0:
        print("{}: {}".format(l, (fl).factor()))
"""
print ("-(2-i)^3 = {}".format(-(2-Fp(u/v))^3))
-(2-i)^3 = 540

2: (z + 1)^2
3: z^2 + 1
5: (z + 2) * (z + 3)
7: z^2 + 1
11: z^2 + 1
13: (z + 5) * (z + 8)
"""

Kf1.<i> = NumberField(x^2+1)

print -1 == i^2
print 2  == (1+i)*(1-i)
print 5  == (2+i)*(2-i)
print 13 == (2+3*i)*(2-3*i)

#F_f = {2:[1+i,1-i], 5:[2+i,2-i], 13:[2+3*i,2-3*i]}
#idx_ideal_f1 = {(2,1+i):9, (2,1-i):10, (5,2+i):11, (5,2-i):12, (13,2+3*i):13, (13,2-3*i):14}
# without the ramified 2: count only (2,1_i) not (2,1-i)
F_f = {2:[1+i], 5:[2+i,2-i], 13:[2+3*i,2-3*i]}
idx_ideal_f1 = {(2,1+i):9, (5,2+i):10, (5,2-i):11, (13,2+3*i):12, (13,2-3*i):13}
idx_unit_1 = 7
idx_unit_i = 8

# rational side 1st, so that we directy get log 2 = 1 (i.e. logs are in basis 2)
F_g = [2,3,5,7,11,13]
idx_ideal_f0 = {2:0, 3:1, 5:2, 7:3, 11:4, 13:5} # index 6 will be for 1/V
idx_lc_f0 = 6

rows = 15 # 16
cols = 14 # 15
mat = [[0 for j in range(cols)] for k in range(rows)]
for k in range(rows):
    mat[k][idx_lc_f0] = 1 # always multiply rational side by 1/V
    
# -i^2 = 1
# vlog(-1) + 2*vlog(i) = 0
row_id = 0
mat[row_id][idx_unit_1] = 1
mat[row_id][idx_unit_i] = 2
mat[row_id][idx_lc_f0] = 0
row_id = 1
# in fact, -i*(1+i) = 1-i
#mat[1][idx_unit_1] = 1
#mat[1][idx_unit_i] = 1
#mat[1][idx_ideal_f1[(2,1+i)]] = 1
#mat[1][idx_ideal_f1[(2,1-i)]] = -1
#mat[1][idx_lc_f0] = 0
#row_id = 2
print("")
for a in range(-A,A+1):
    for b in [b for b in range(1,A+1) if gcd(a,b) == 1]:
        nf1 = ZZ(a^2+b^2)
        nf0 = ZZ(v*a+u*b) #a-b*x, a*v-u*(-b) = a*v-b*u#
        if nf1 == 0 or nf0 == 0:
            continue
        fac_f1 = nf1.factor()
        fac_f0 = nf0.factor()
        # if fac_f is +/-1, then fac_f = [+/-1]
        # otherwise, fac_f = [(q1,e1), ..., (qi,ei)]
        #print("({},{}): {} = {} (f), {} = {} (g)".format(a,b, nf, fac_f, ng, fac_g))
        smooth_f1 = True
        for l,ei in fac_f1:
            if l > B:
                smooth_f1 = False
                break
        smooth_f0 = True
        for l,_ in fac_f0:
            if l > B:
                smooth_f0 = False
                break
        if smooth_f1 and smooth_f0:
            str_f1 = ""
            prod_id_f1 = Kf1(1)
            for (l,ei) in fac_f1:
                # pb if l=2: cannot make a difference between the two ideals, they are both (x+1) mod 2.
                Fl = GF(l)
                Flz.<z> = Fl[]
                ideal_f1 = gcd(Flz(f1), Flz(a+b*x)) # or a-bx
                for id_x in F_f[l]:
                    if gcd(Flz(id_x.polynomial()), ideal_f1) > 1:
                        str_f1 += "("+str(l)+","+str(id_x)+")"
                        if ei>1: str_f1 += "^"+str(ei)
                        prod_id_f1 *= id_x^ei
                        mat[row_id][idx_ideal_f1[(l,id_x)]] = -ei # everything negative (-) on the algebraic side
                        break
            unit_f1 = (a+b*i)/prod_id_f1 # unit
            if unit_f1 == i or unit_f1 == -i:
                mat[row_id][idx_unit_i] = -1
            if unit_f1 == -1 or unit_f1 == -i:
                mat[row_id][idx_unit_1] = -1
            
            for (l,ei) in fac_f0:
                mat[row_id][idx_ideal_f0[l]] = ei
            if nf0 < 0: # multiply by -1
                mat[row_id][idx_unit_1] = (mat[row_id][idx_unit_1] + 1) % 2
            # a,b a*V-b*U=fac a^2+b^2=fac fac_Zi
            print("{:3},{:2} & {: 4} & {:13s} & {:3d} & {:15s} & {:33s} & {} \\\\".format(a,b,nf0,pretty_print_tex_product(nf0,fac_f0), nf1,pretty_print_tex_product(nf1,fac_f1), str_f1, unit_f1))

            
            row_id += 1
print("")

M2 = Matrix(Fr, rows, cols, mat)
M1 = Matrix(ZZ, rows, cols, mat)
print M1
print "\n"
print M2
print M2.rank()
print M2.right_kernel()

print("check matrix M2")
i_p = Fp(u/v)
#col_a = [-1, i_p, (1+i_p), (1-i_p), (2+i_p), (2-i_p), (2+3*i_p), (2-3*i_p)]
col_a = [-1, i_p, (1+i_p), (2+i_p), (2-i_p), (2+3*i_p), (2-3*i_p)]
col_r = [2,  3,  5,  7, 11, 13, 1/v]
c = 0
for row in M2.rows():
    pr_a = Fp(1)
    pr_r = Fp(1)
    for j in range(len(col_r)):
        ej = ZZ(row[j])
        pr_r *= Fp(col_r[j])^ej
    for j in range(len(col_a)):
        ej = ZZ(row[j+len(col_r)])
        pr_a *= Fp(col_a[j])^(-ej)
    print("{:2}: {:3}, {:3}, ({:3}/{:3})^4={}".format(c, pr_r, pr_a, pr_r, pr_a, (pr_r/pr_a)^4))
    c += 1

"""
-17,19 &   -7 & -7            & 650 & 2\cd 5^2\cd 13  & (2,i+1)(5,i+2)^2(13,-3*i+2) &  i \\
-11, 2 & -231 & -3\cd 7\cd 11 & 125 & 5^3             & (5,i+2)^3                   &  i \\
 -6,17 &  224 & 2^5\cd 7      & 325 & 5^2\cd 13       & (5,i+2)^2(13,3*i+2)         &  1 \\
 -4, 7 &   54 & 2\cd 3^3      &  65 & 5\cd 13         & (5,-i+2)(13,3*i+2)          &  i \\
 -3, 4 &   13 & 13            &  25 & 5^2             & (5,-i+2)^2                  & -1 \\
 -2, 1 &  -28 & -2^2\cd 7     &   5 & 5               & (5,-i+2)                    & -1 \\
 -2, 3 &   16 & 2^4           &  13 & 13              & (13,-3*i+2)                 & -1 \\
 -2,11 &  192 & 2^6\cd 3      & 125 & 5^3             & (5,-i+2)^3                  & -1 \\
 -1, 1 &   -3 & -3            &   2 & 2               & (2,i+1)                     &  i \\
  0, 1 &   22 & 2\cd 11       &   1 & 1               &                             &  i \\
  1, 3 &   91 & 7\cd 13       &  10 & 2\cd 5          & (2,i+1)(5,i+2)              &  1 \\
  1, 5 &  135 & 3^3\cd 5      &  26 & 2\cd 13         & (2,i+1)(13,-3*i+2)          &  i \\
  2, 1 &   72 & 2^3\cd 3^2    &   5 & 5               & (5,i+2)                     &  1 \\
  5, 1 &  147 & 3\cd 7^2      &  26 & 2\cd 13         & (2,i+1)(13,3*i+2)           & -i \\

(-17+19*i)== -(1-i)*(2+i)^2*(2-3*i)
(-11+2*i) == i*(2+i)^3
(-6+17*i) == (2+i)^2*(2+3*i)
(-4+7*i)  == i*(2-i)*(2+3*i)
(-3+4*i)  == -(2-i)^2
(-2+i)    == -(2-i)
(-2+3*i)  == -(2-3*i)
(-2+11*i) == -(2-i)^3
(-1+i)    == -(1-i)
i         == i
(1+3*i)   == (1+3*i)
(1+5*i)   == -(1-i)*(2-3*i)
(2+i)     == (2+i)
(5 +i)    == -i*(1+i)*(2+3*i)
"""
test_fac = [
    ((-17+19*i), -  (1-i)*(2+i)^2*(2-3*i) ,   -7, [-1,             7       ]),
    ((-11+2*i) ,  i*      (2+i)^3         , -231, [-1,      3,     7, 11   ]),
    ((-6+17*i) ,          (2+i)^2*(2+3*i) ,  224, [    2^5,        7       ]),
    ((-4+7*i)  ,  i*      (2-i)  *(2+3*i) ,   54, [    2,   3^3            ]),
    ((-3+4*i)  , -        (2-i)^2         ,   13, [                      13]),
    ((-2+i)    , -        (2-i)           ,  -28, [-1, 2^2,        7       ]),
    ((-2+3*i)  , -                (2-3*i) ,   16, [    2^4                 ]),
    ((-2+11*i) , -        (2-i)^3         ,  192, [    2^6, 3              ]),
    ((-1+i)    , -  (1-i)                 ,   -3, [-1,      3              ]),
    (i         ,  i                       ,   22, [    2,             11   ]),
    ((1+3*i)   ,    (1+i)*(2+i)           ,   91, [                7,    13]),
    ((1+5*i)   , -  (1-i)*        (2-3*i) ,  135, [         3^3, 5         ]),
    ((2+i)     ,          (2+i)           ,   72, [    2^3, 3^2            ]),
    ((5 +i)    , -i*(1+i)*        (2+3*i) ,  147, [         3,     7^2     ])
]

for (ab, fac_f, ng, fac_g) in test_fac:
    print (ab - fac_f)
    print ng == prod(fac_g)

# cancel out the factor (-1)
test_fac = [
    (-(-17+19*i),    (1-i)*(2+i)^2*(2-3*i)  ,    7, [             7       ]),
    (-(-11+2*i) , -i*      (2+i)^3          ,  231, [      3,     7, 11   ]),
    ( (-6+17*i) ,          (2+i)^2*(2+3*i)  ,  224, [ 2^5,        7       ]),
    ( (-4+7*i)  ,  i*      (2-i)  *(2+3*i)  ,   54, [ 2,   3^3            ]),
    ( (-3+4*i)  , -        (2-i)^2          ,   13, [                   13]),
    (-(-2+i)    ,          (2-i)            ,   28, [ 2^2,        7       ]),
    ( (-2+3*i)  , -                (2-3*i)  ,   16, [ 2^4                 ]),
    ( (-2+11*i) , -        (2-i)^3          ,  192, [ 2^6, 3              ]),
    (-(-1+i)    ,    (1-i)                  ,    3, [      3              ]),
    ( i         ,  i                        ,   22, [ 2,             11   ]),
    ( (1+3*i)   ,    (1+i)*(2+i)            ,   91, [             7,    13]),
    ( (1+5*i)   , -  (1-i)*        (2-3*i)  ,  135, [      3^3, 5         ]),
    ( (2+i)     ,          (2+i)            ,   72, [ 2^3, 3^2            ]),
    ( (5 +i)    , -i*(1+i)*        (2+3*i)  ,  147, [      3,     7^2     ]),
    ( (22+19*i) , -        (2-i)*  (2-3*i)^2,  968, [ 2^3,           11^2 ])
]
# matrix of exponents
Zq = Zmod(p-1)
M = Matrix(Fr, 15, 15, \
           # 2  3  5  7 11 13 1/v    -1 i (1+i) (1-i) (2+i) (2-i) (2+3i) (2-3i) 
           [0, 0, 0, 0, 0, 0, 0,     1, 2,   0,    0,    0,    0,     0,     0,  \
            0, 0, 0, 1, 0, 0, 1,     0, 0,   0,    1,    2,    0,     0,     1,  \
            0, 1, 0, 1, 1, 0, 1,     1, 1,   0,    0,    3,    0,     0,     0,  \
            5, 0, 0, 1, 0, 0, 1,     0, 0,   0,    0,    2,    0,     1,     0,  \
            1, 3, 0, 0, 0, 0, 1,     0, 1,   0,    0,    0,    1,     1,     0,  \
            0, 0, 0, 0, 0, 1, 1,     1, 0,   0,    0,    0,    2,     0,     0,  \
            2, 0, 0, 1, 0, 0, 1,     0, 0,   0,    0,    0,    1,     0,     0,  \
            4, 0, 0, 0, 0, 0, 1,     1, 0,   0,    0,    0,    0,     0,     1,  \
            6, 1, 0, 0, 0, 0, 1,     1, 0,   0,    0,    0,    3,     0,     0,  \
            0, 1, 0, 0, 0, 0, 1,     0, 0,   0,    1,    0,    0,     0,     0,  \
            1, 0, 0, 0, 1, 0, 1,     0, 1,   0,    0,    0,    0,     0,     0,  \
            0, 0, 0, 1, 0, 1, 1,     0, 0,   1,    0,    1,    0,     0,     0,  \
            0, 3, 1, 0, 0, 0, 1,     1, 0,   0,    1,    0,    0,     0,     1,  \
            3, 2, 0, 0, 0, 0, 1,     0, 0,   0,    0,    1,    0,     0,     0,  \
            0, 1, 0, 2, 0, 0, 1,     1, 1,   1,    0,    0,    0,     1,     0, \
            #3, 0, 0, 0, 2, 0, 1,     1, 0,   0,    0,    0,    1,     0,     2,  \
           ])

print("check matrix M")
i_p = Fp(u/v)
col_a = [-1, i_p, (1+i_p), (1-i_p), (2+i_p), (2-i_p), (2+3*i_p), (2-3*i_p)]
col_r = [2,  3,  5,  7, 11, 13, 1/v]
c = 0
for row in M.rows():
    pr_a = Fp(1)
    pr_r = Fp(1)
    for j in range(len(col_r)):
        ej = ZZ(row[j])
        pr_r *= Fp(col_r[j])^ej
    for j in range(len(col_a)):
        ej = ZZ(row[j+len(col_r)])
        pr_a *= Fp(col_a[j])^(ej)
    #print("{}: {}, {}".format(c, pr_a, pr_r))
    print("{:2}: {:3}, {:3}, ({:3}/{:3})^4={}".format(c, pr_r, pr_a, pr_r, pr_a, (pr_r/pr_a)^4))
    c += 1
    
print M.right_kernel()
# vlog (1+i) = vlog 1-i because -i(1+i) = 1-i, and vlog(-1) = vlog(i) = 0

#14
#Vector space of degree 15 and dimension 1 over Finite Field of size 277
#Basis matrix: (with -1 for algebraic side)
#[  1 219  40  34  79 269 197   0   0 139 139  84 233  68 201]
#Basis matrix: (no -1 for algebraic side)
#[  1 219  40  34  79 269 197   0   0 138 138 193  44 209  76]

log_2 = 1
log_3 = 219
log_5 = 40
log_7 = 37
log_11 = 79
log_13 = 269
log_v_ = 197

target = 314
t = 1
nr = ZZ(target*2^t % p)
fac_n = nr.factor()
smooth_n = True
for (l,_) in fac_n:
    if l > B:
        smooth_n = False
        break
while not smooth_n:
    t += 1
    nr = ZZ(target*2^t % p)
    fac_n = nr.factor()
    smooth_n = True
    for (l,_) in fac_n:
        if l > B:
            smooth_n = False
            break
print t, nr, fac_n

(2^2*314 % p) == 3*7^2
inv_log_2 = 138
log_target = (log_3 + 2*log_7 - 2*log_2)*inv_log_2 % ((p-1)//4)

Fp(2)^(839) == 314

print("Individual log with rational reconstruction")
target = 314
t = -1
smooth_target_found = False
while not smooth_target_found:
    t += 1
    nr = ZZ(target*2**t % p)
    N = target.rational_reconstruction(p)
    num = N.numer()
    den = N.denom()
    smooth_target_found = is_smooth(num,B) and is_smooth(den,B)
print("g^{}*{} = {}/{}".format(t,target,num,den))
# -20/7 = 314 mod p
